'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var React = require('react');
var React__default = _interopDefault(React);

var simpleIsEqual = function simpleIsEqual(a, b) {
  return a === b;
};

function memoizeOne (resultFn) {
  var isEqual = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : simpleIsEqual;

  var lastThis = void 0;
  var lastArgs = [];
  var lastResult = void 0;
  var calledOnce = false;

  var isNewArgEqualToLast = function isNewArgEqualToLast(newArg, index) {
    return isEqual(newArg, lastArgs[index]);
  };

  var result = function result() {
    for (var _len = arguments.length, newArgs = Array(_len), _key = 0; _key < _len; _key++) {
      newArgs[_key] = arguments[_key];
    }

    if (calledOnce && lastThis === this && newArgs.length === lastArgs.length && newArgs.every(isNewArgEqualToLast)) {
      return lastResult;
    }

    calledOnce = true;
    lastThis = this;
    lastArgs = newArgs;
    lastResult = resultFn.apply(this, newArgs);
    return lastResult;
  };

  return result;
}

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
};

var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();

var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};

var inherits = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};

var possibleConstructorReturn = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};

var slicedToArray = function () {
  function sliceIterator(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"]) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  return function (arr, i) {
    if (Array.isArray(arr)) {
      return arr;
    } else if (Symbol.iterator in Object(arr)) {
      return sliceIterator(arr, i);
    } else {
      throw new TypeError("Invalid attempt to destructure non-iterable instance");
    }
  };
}();

var IS_SCROLLING_DEBOUNCE_INTERVAL = 150;

function createGridComponent(_ref) {
  var _class, _temp;

  var getColumnOffset = _ref.getColumnOffset,
      getColumnStartIndexForOffset = _ref.getColumnStartIndexForOffset,
      getColumnStopIndexForStartIndex = _ref.getColumnStopIndexForStartIndex,
      getColumnWidth = _ref.getColumnWidth,
      getEstimatedTotalHeight = _ref.getEstimatedTotalHeight,
      getEstimatedTotalWidth = _ref.getEstimatedTotalWidth,
      getOffsetForColumnAndAlignment = _ref.getOffsetForColumnAndAlignment,
      getOffsetForRowAndAlignment = _ref.getOffsetForRowAndAlignment,
      getRowHeight = _ref.getRowHeight,
      getRowOffset = _ref.getRowOffset,
      getRowStartIndexForOffset = _ref.getRowStartIndexForOffset,
      getRowStopIndexForStartIndex = _ref.getRowStopIndexForStartIndex,
      initInstanceProps = _ref.initInstanceProps,
      validateProps = _ref.validateProps;

  return _temp = _class = function (_PureComponent) {
    inherits(Grid, _PureComponent);

    function Grid(props) {
      classCallCheck(this, Grid);

      var _this = possibleConstructorReturn(this, (Grid.__proto__ || Object.getPrototypeOf(Grid)).call(this, props));

      _this._cellStyleCache = {};
      _this._resetIsScrollingTimeoutId = null;
      _this.state = {
        isScrolling: false,
        horizontalScrollDirection: 'forward',
        scrollLeft: typeof _this.props.defaultScrollLeft === 'number' ? _this.props.defaultScrollLeft : 0,
        scrollTop: typeof _this.props.defaultScrollTop === 'number' ? _this.props.defaultScrollTop : 0,
        verticalScrollDirection: 'forward'
      };
      _this._callOnItemsRendered = memoizeOne(function (overscanColumnStartIndex, overscanColumnStopIndex, overscanRowStartIndex, overscanRowStopIndex, visibleColumnStartIndex, visibleColumnStopIndex, visibleRowStartIndex, visibleRowStopIndex) {
        return _this.props.onItemsRendered({
          overscanColumnStartIndex: overscanColumnStartIndex,
          overscanColumnStopIndex: overscanColumnStopIndex,
          overscanRowStartIndex: overscanRowStartIndex,
          overscanRowStopIndex: overscanRowStopIndex,
          visibleColumnStartIndex: visibleColumnStartIndex,
          visibleColumnStopIndex: visibleColumnStopIndex,
          visibleRowStartIndex: visibleRowStartIndex,
          visibleRowStopIndex: visibleRowStopIndex
        });
      });
      _this._callOnScroll = memoizeOne(function (scrollLeft, scrollTop, horizontalScrollDirection, verticalScrollDirection) {
        return _this.props.onScroll({
          horizontalScrollDirection: horizontalScrollDirection,
          scrollLeft: scrollLeft,
          scrollTop: scrollTop,
          verticalScrollDirection: verticalScrollDirection
        });
      });

      _this._getCellStyle = function (rowIndex, columnIndex) {
        var key = rowIndex + ':' + columnIndex;

        var style = void 0;
        if (_this._cellStyleCache.hasOwnProperty(key)) {
          style = _this._cellStyleCache[key];
        } else {
          _this._cellStyleCache[key] = style = {
            position: 'absolute',
            left: getColumnOffset(_this.props, columnIndex, _this._instanceProps),
            top: getRowOffset(_this.props, rowIndex, _this._instanceProps),
            height: getRowHeight(_this.props, rowIndex, _this._instanceProps),
            width: getColumnWidth(_this.props, columnIndex, _this._instanceProps)
          };
        }

        return style;
      };

      _this._onScroll = function (event) {
        var _event$currentTarget = event.currentTarget,
            scrollLeft = _event$currentTarget.scrollLeft,
            scrollTop = _event$currentTarget.scrollTop;

        _this.setState(function (prevState) {
          return {
            isScrolling: true,
            horizontalScrollDirection: prevState.scrollLeft < scrollLeft ? 'forward' : 'backward',
            scrollLeft: scrollLeft,
            scrollTop: scrollTop,
            verticalScrollDirection: prevState.scrollTop < scrollTop ? 'forward' : 'backward'
          };
        }, _this._resetIsScrollingDebounced);
      };

      _this._scrollingContainerRef = function (ref) {
        _this._scrollingContainer = ref;
      };

      _this._renderFunction = function (params) {
        return _this.props.children(params);
      };

      _this._resetIsScrollingDebounced = function () {
        if (_this._resetIsScrollingTimeoutId !== null) {
          clearTimeout(_this._resetIsScrollingTimeoutId);
        }

        _this._resetIsScrollingTimeoutId = setTimeout(_this._resetIsScrolling, IS_SCROLLING_DEBOUNCE_INTERVAL);
      };

      _this._resetIsScrollingDebounced = function () {
        if (_this._resetIsScrollingTimeoutId !== null) {
          clearTimeout(_this._resetIsScrollingTimeoutId);
        }

        _this._resetIsScrollingTimeoutId = setTimeout(_this._resetIsScrolling, IS_SCROLLING_DEBOUNCE_INTERVAL);
      };

      _this._resetIsScrolling = function () {
        _this._resetIsScrollingTimeoutId = null;

        _this.setState({ isScrolling: false }, function () {
          // Clear style cache after state update has been committed.
          // This way we don't break pure sCU for cells that don't use isScrolling param.
          _this._cellStyleCache = {};
        });
      };

      _this._instanceProps = initInstanceProps(props, _this);
      return _this;
    }

    createClass(Grid, [{
      key: 'scrollTo',
      value: function scrollTo(_ref2) {
        var scrollLeft = _ref2.scrollLeft,
            scrollTop = _ref2.scrollTop;

        if (this._scrollingContainer != null) {
          this._scrollingContainer.scrollLeft = scrollLeft;
          this._scrollingContainer.scrollTop = scrollTop;
        }

        if (process.env.NODE_ENV === 'test') {
          // Setting scroll offset doesn't fire the onScroll callback for react-test-renderer.
          // This test-only code makes it easier to test simualted scrolling behavior.
          // It should be stripped out of any non-test code.
          this._onScroll({ currentTarget: { scrollLeft: scrollLeft, scrollTop: scrollTop } });
        }
      }
    }, {
      key: 'scrollToItem',
      value: function scrollToItem(_ref3) {
        var _ref3$align = _ref3.align,
            align = _ref3$align === undefined ? 'auto' : _ref3$align,
            columnIndex = _ref3.columnIndex,
            rowIndex = _ref3.rowIndex;
        var _state = this.state,
            scrollLeft = _state.scrollLeft,
            scrollTop = _state.scrollTop;


        this.scrollTo({
          scrollLeft: getOffsetForColumnAndAlignment(this.props, columnIndex, align, scrollLeft, this._instanceProps),
          scrollTop: getOffsetForRowAndAlignment(this.props, rowIndex, align, scrollTop, this._instanceProps)
        });
      }
    }, {
      key: 'componentDidMount',
      value: function componentDidMount() {
        var _props = this.props,
            defaultScrollLeft = _props.defaultScrollLeft,
            defaultScrollTop = _props.defaultScrollTop;


        if (typeof defaultScrollLeft === 'number' && this._scrollingContainer != null) {
          this._scrollingContainer.scrollLeft = defaultScrollLeft;
        }
        if (typeof defaultScrollTop === 'number' && this._scrollingContainer != null) {
          this._scrollingContainer.scrollTop = defaultScrollTop;
        }

        this._callPropsCallbacks();
      }
    }, {
      key: 'componentDidUpdate',
      value: function componentDidUpdate() {
        this._callPropsCallbacks();
      }
    }, {
      key: 'componentWillUnmount',
      value: function componentWillUnmount() {
        if (this._resetIsScrollingTimeoutId !== null) {
          clearTimeout(this._resetIsScrollingTimeoutId);
        }
      }
    }, {
      key: 'render',
      value: function render() {
        var _props2 = this.props,
            className = _props2.className,
            columnCount = _props2.columnCount,
            height = _props2.height,
            rowCount = _props2.rowCount,
            style = _props2.style,
            useIsScrolling = _props2.useIsScrolling,
            width = _props2.width;
        var isScrolling = this.state.isScrolling;


        var estimatedTotalHeight = getEstimatedTotalHeight(this.props, this._instanceProps);
        var estimatedTotalWidth = getEstimatedTotalWidth(this.props, this._instanceProps);

        var _getHorizontalRangeTo = this._getHorizontalRangeToRender(),
            _getHorizontalRangeTo2 = slicedToArray(_getHorizontalRangeTo, 2),
            columnStartIndex = _getHorizontalRangeTo2[0],
            columnStopIndex = _getHorizontalRangeTo2[1];

        var _getVerticalRangeToRe = this._getVerticalRangeToRender(),
            _getVerticalRangeToRe2 = slicedToArray(_getVerticalRangeToRe, 2),
            rowStartIndex = _getVerticalRangeToRe2[0],
            rowStopIndex = _getVerticalRangeToRe2[1];

        return React__default.createElement(
          'div',
          {
            className: className,
            ref: this._scrollingContainerRef,
            style: _extends({
              position: 'relative',
              height: height,
              width: width,
              overflow: 'auto',
              WebkitOverflowScrolling: 'touch',
              willChange: 'transform'
            }, style),
            onScroll: this._onScroll
          },
          React__default.createElement(
            'div',
            {
              style: {
                height: estimatedTotalHeight,
                overflow: 'hidden',
                pointerEvents: isScrolling ? 'none' : '',
                width: estimatedTotalWidth
              }
            },
            columnCount > 0 && rowCount > 0 && React__default.createElement(GridItems, {
              columnStartIndex: columnStartIndex,
              columnStopIndex: columnStopIndex,
              getCellStyle: this._getCellStyle,
              isScrolling: useIsScrolling ? isScrolling : undefined,
              renderFunction: this._renderFunction,
              rowStartIndex: rowStartIndex,
              rowStopIndex: rowStopIndex
            })
          )
        );
      }
    }, {
      key: '_callPropsCallbacks',
      value: function _callPropsCallbacks() {
        if (typeof this.props.onItemsRendered === 'function') {
          var _getHorizontalRangeTo3 = this._getHorizontalRangeToRender(),
              _getHorizontalRangeTo4 = slicedToArray(_getHorizontalRangeTo3, 4),
              _overscanColumnStartIndex = _getHorizontalRangeTo4[0],
              _overscanColumnStopIndex = _getHorizontalRangeTo4[1],
              _visibleColumnStartIndex = _getHorizontalRangeTo4[2],
              _visibleColumnStopIndex = _getHorizontalRangeTo4[3];

          var _getVerticalRangeToRe3 = this._getVerticalRangeToRender(),
              _getVerticalRangeToRe4 = slicedToArray(_getVerticalRangeToRe3, 4),
              _overscanRowStartIndex = _getVerticalRangeToRe4[0],
              _overscanRowStopIndex = _getVerticalRangeToRe4[1],
              _visibleRowStartIndex = _getVerticalRangeToRe4[2],
              _visibleRowStopIndex = _getVerticalRangeToRe4[3];

          this._callOnItemsRendered(_overscanColumnStartIndex, _overscanColumnStopIndex, _overscanRowStartIndex, _overscanRowStopIndex, _visibleColumnStartIndex, _visibleColumnStopIndex, _visibleRowStartIndex, _visibleRowStopIndex);
        }

        if (typeof this.props.onScroll === 'function') {
          var _state2 = this.state,
              _horizontalScrollDirection = _state2.horizontalScrollDirection,
              _scrollLeft = _state2.scrollLeft,
              _scrollTop = _state2.scrollTop,
              _verticalScrollDirection = _state2.verticalScrollDirection;

          this._callOnScroll(_scrollLeft, _scrollTop, _horizontalScrollDirection, _verticalScrollDirection);
        }
      }

      // Lazily create and cache cell styles while scrolling,
      // So that pure component sCU will prevent re-renders.

    }, {
      key: '_getHorizontalRangeToRender',
      value: function _getHorizontalRangeToRender() {
        var _props3 = this.props,
            columnCount = _props3.columnCount,
            overscanCount = _props3.overscanCount;
        var _state3 = this.state,
            horizontalScrollDirection = _state3.horizontalScrollDirection,
            scrollLeft = _state3.scrollLeft;


        var startIndex = getColumnStartIndexForOffset(this.props, scrollLeft, this._instanceProps);
        var stopIndex = getColumnStopIndexForStartIndex(this.props, startIndex, scrollLeft, this._instanceProps);

        // Overscan by one cell in each direction so that tab/focus works.
        // If there isn't at least one extra cell, tab loops back around.
        var overscanBackward = horizontalScrollDirection === 'backward' ? Math.max(1, overscanCount) : 1;
        var overscanForward = horizontalScrollDirection === 'forward' ? Math.max(1, overscanCount) : 1;

        return [Math.max(0, startIndex - overscanBackward), Math.max(0, Math.min(columnCount - 1, stopIndex + overscanForward)), startIndex, stopIndex];
      }
    }, {
      key: '_getVerticalRangeToRender',
      value: function _getVerticalRangeToRender() {
        var _props4 = this.props,
            rowCount = _props4.rowCount,
            overscanCount = _props4.overscanCount;
        var _state4 = this.state,
            verticalScrollDirection = _state4.verticalScrollDirection,
            scrollTop = _state4.scrollTop;


        var startIndex = getRowStartIndexForOffset(this.props, scrollTop, this._instanceProps);
        var stopIndex = getRowStopIndexForStartIndex(this.props, startIndex, scrollTop, this._instanceProps);

        // Overscan by one cell in each direction so that tab/focus works.
        // If there isn't at least one extra cell, tab loops back around.
        var overscanBackward = verticalScrollDirection === 'backward' ? Math.max(1, overscanCount) : 1;
        var overscanForward = verticalScrollDirection === 'forward' ? Math.max(1, overscanCount) : 1;

        return [Math.max(0, startIndex - overscanBackward), Math.max(0, Math.min(rowCount - 1, stopIndex + overscanForward)), startIndex, stopIndex];
      }

      // Facade for the user-provided children function.
      // This adds the overhead of an additional method call,
      // But has hte benefit of not breaking pure sCU checks,
      // Allowing List to avoid re-rendering cells until indices change.

    }], [{
      key: 'getDerivedStateFromProps',
      value: function getDerivedStateFromProps(nextProps, prevState) {
        validateSharedProps(nextProps);
        validateProps(nextProps);
        return null;
      }
    }]);
    return Grid;
  }(React.PureComponent), _class.defaultProps = {
    overscanCount: 1,
    useIsScrolling: false
  }, _temp;
}

var GridItems = function (_PureComponent2) {
  inherits(GridItems, _PureComponent2);

  function GridItems() {
    classCallCheck(this, GridItems);
    return possibleConstructorReturn(this, (GridItems.__proto__ || Object.getPrototypeOf(GridItems)).apply(this, arguments));
  }

  createClass(GridItems, [{
    key: 'render',
    value: function render() {
      var _props5 = this.props,
          columnStartIndex = _props5.columnStartIndex,
          columnStopIndex = _props5.columnStopIndex,
          getCellStyle = _props5.getCellStyle,
          isScrolling = _props5.isScrolling,
          renderFunction = _props5.renderFunction,
          rowStartIndex = _props5.rowStartIndex,
          rowStopIndex = _props5.rowStopIndex;


      var cells = [];

      for (var _rowIndex = rowStartIndex; _rowIndex <= rowStopIndex; _rowIndex++) {
        for (var _columnIndex = columnStartIndex; _columnIndex <= columnStopIndex; _columnIndex++) {
          var _key = _rowIndex + ':' + _columnIndex;
          var _style = getCellStyle(_rowIndex, _columnIndex);

          cells.push(renderFunction({
            columnIndex: _columnIndex,
            key: _key,
            isScrolling: isScrolling,
            rowIndex: _rowIndex,
            style: _style
          }));
        }
      }

      return cells;
    }
  }]);
  return GridItems;
}(React.PureComponent);

var validateSharedProps = function validateSharedProps(_ref4) {
  var children = _ref4.children,
      height = _ref4.height,
      width = _ref4.width;

  if (process.env.NODE_ENV !== 'production') {
    if (typeof children !== 'function') {
      throw Error('An invalid "children" prop has been specified. ' + 'Value should be a function that creates a React element. ' + ('"' + (children === null ? 'null' : typeof children === 'undefined' ? 'undefined' : _typeof(children)) + '" was specified.'));
    }

    if (typeof width !== 'number') {
      throw Error('An invalid "width" prop has been specified. ' + 'Grids must specify a number for width. ' + ('"' + (width === null ? 'null' : typeof width === 'undefined' ? 'undefined' : _typeof(width)) + '" was specified.'));
    }

    if (typeof height !== 'number') {
      throw Error('An invalid "height" prop has been specified. ' + 'Grids must specify a number for height. ' + ('"' + (height === null ? 'null' : typeof height === 'undefined' ? 'undefined' : _typeof(height)) + '" was specified.'));
    }
  }
};

var DEFAULT_ESTIMATED_CELL_SIZE = 50;

var getCellMetadata = function getCellMetadata(cellType, props, index, instanceProps) {
  var cellMetadataMap = void 0,
      cellSize = void 0,
      lastMeasuredIndex = void 0;
  if (cellType === 'column') {
    cellMetadataMap = instanceProps.columnMetadataMap;
    cellSize = props.columnWidth;
    lastMeasuredIndex = instanceProps.lastMeasuredColumnIndex;
  } else {
    cellMetadataMap = instanceProps.rowMetadataMap;
    cellSize = props.rowHeight;
    lastMeasuredIndex = instanceProps.lastMeasuredRowIndex;
  }

  if (index > lastMeasuredIndex) {
    var _offset = 0;
    if (lastMeasuredIndex >= 0) {
      var cellMetadata = cellMetadataMap[lastMeasuredIndex];
      _offset = cellMetadata.offset + cellMetadata.size;
    }

    for (var i = lastMeasuredIndex + 1; i <= index; i++) {
      var _size = cellSize(i);

      cellMetadataMap[i] = {
        offset: _offset,
        size: _size
      };

      _offset += _size;
    }

    if (cellType === 'column') {
      instanceProps.lastMeasuredColumnIndex = index;
    } else {
      instanceProps.lastMeasuredRowIndex = index;
    }
  }

  return cellMetadataMap[index];
};

var findNearestCell = function findNearestCell(cellType, props, instanceProps, offset) {
  var cellMetadataMap = void 0,
      lastMeasuredIndex = void 0;
  if (cellType === 'column') {
    cellMetadataMap = instanceProps.columnMetadataMap;
    lastMeasuredIndex = instanceProps.lastMeasuredColumnIndex;
  } else {
    cellMetadataMap = instanceProps.rowMetadataMap;
    lastMeasuredIndex = instanceProps.lastMeasuredRowIndex;
  }

  var lastMeasuredCellOffset = lastMeasuredIndex > 0 ? cellMetadataMap[lastMeasuredIndex].offset : 0;

  if (lastMeasuredCellOffset >= offset) {
    // If we've already measured cells within this range just use a binary search as it's faster.
    return findNearestCellBinarySearch(cellType, props, instanceProps, lastMeasuredIndex, 0, offset);
  } else {
    // If we haven't yet measured this high, fallback to an exponential search with an inner binary search.
    // The exponential search avoids pre-computing sizes for the full set of cells as a binary search would.
    // The overall complexity for this approach is O(log n).
    return findNearestCellExponentialSearch(cellType, props, instanceProps, lastMeasuredIndex, offset);
  }
};

var findNearestCellBinarySearch = function findNearestCellBinarySearch(cellType, props, instanceProps, high, low, offset) {
  while (low <= high) {
    var middle = low + Math.floor((high - low) / 2);
    var currentOffset = getCellMetadata(cellType, props, middle, instanceProps).offset;

    if (currentOffset === offset) {
      return middle;
    } else if (currentOffset < offset) {
      low = middle + 1;
    } else if (currentOffset > offset) {
      high = middle - 1;
    }
  }

  if (low > 0) {
    return low - 1;
  } else {
    return 0;
  }
};

var findNearestCellExponentialSearch = function findNearestCellExponentialSearch(cellType, props, instanceProps, index, offset) {
  var count = cellType === 'column' ? props.columnCount : props.rowCount;
  var interval = 1;

  while (index < count && getCellMetadata(cellType, props, index, instanceProps).offset < offset) {
    index += interval;
    interval *= 2;
  }

  return findNearestCellBinarySearch(cellType, props, instanceProps, Math.min(index, count - 1), Math.floor(index / 2), offset);
};

var getOffsetForIndexAndAlignment = function getOffsetForIndexAndAlignment(cellType, props, index, align, scrollOffset, instanceProps) {
  var size = cellType === 'column' ? props.width : props.height;
  var cellMetadata = getCellMetadata(cellType, props, index, instanceProps);
  var maxOffset = cellMetadata.offset;
  var minOffset = cellMetadata.offset - size + cellMetadata.size;

  switch (align) {
    case 'start':
      return maxOffset;
    case 'end':
      return minOffset;
    case 'center':
      return Math.round(minOffset + (maxOffset - minOffset) / 2);
    case 'auto':
    default:
      if (scrollOffset >= minOffset && scrollOffset <= maxOffset) {
        return scrollOffset;
      } else if (scrollOffset - minOffset < maxOffset - scrollOffset) {
        return minOffset;
      } else {
        return maxOffset;
      }
  }
};

var VariableSizeGrid = createGridComponent({
  getColumnOffset: function getColumnOffset(props, index, instanceProps) {
    return getCellMetadata('column', props, index, instanceProps).offset;
  },

  getColumnStartIndexForOffset: function getColumnStartIndexForOffset(props, scrollLeft, instanceProps) {
    return findNearestCell('column', props, instanceProps, scrollLeft);
  },

  getColumnStopIndexForStartIndex: function getColumnStopIndexForStartIndex(props, startIndex, scrollLeft, instanceProps) {
    var columnCount = props.columnCount,
        width = props.width;


    var cellMetadata = getCellMetadata('column', props, startIndex, instanceProps);
    var maxOffset = scrollLeft + width;

    var offset = cellMetadata.offset + cellMetadata.size;
    var stopIndex = startIndex;

    while (stopIndex < columnCount - 1 && offset < maxOffset) {
      stopIndex++;
      offset += getCellMetadata('column', props, stopIndex, instanceProps).size;
    }

    return stopIndex;
  },

  getColumnWidth: function getColumnWidth(props, index, instanceProps) {
    return instanceProps.columnMetadataMap[index].size;
  },

  getEstimatedTotalHeight: function getEstimatedTotalHeight(_ref, _ref2) {
    var rowCount = _ref.rowCount;
    var rowMetadataMap = _ref2.rowMetadataMap,
        estimatedRowHeight = _ref2.estimatedRowHeight,
        lastMeasuredRowIndex = _ref2.lastMeasuredRowIndex;

    var totalSizeOfMeasuredRows = 0;

    if (lastMeasuredRowIndex >= 0) {
      var cellMetadata = rowMetadataMap[lastMeasuredRowIndex];
      totalSizeOfMeasuredRows = cellMetadata.offset + cellMetadata.size;
    }

    var numUnmeasuredCells = rowCount - lastMeasuredRowIndex - 1;
    var totalSizeOfUnmeasuredCells = numUnmeasuredCells * estimatedRowHeight;

    return totalSizeOfMeasuredRows + totalSizeOfUnmeasuredCells;
  },

  getEstimatedTotalWidth: function getEstimatedTotalWidth(_ref3, _ref4) {
    var columnCount = _ref3.columnCount;
    var columnMetadataMap = _ref4.columnMetadataMap,
        estimatedColumnWidth = _ref4.estimatedColumnWidth,
        lastMeasuredColumnIndex = _ref4.lastMeasuredColumnIndex;

    var totalSizeOfMeasuredRows = 0;

    if (lastMeasuredColumnIndex >= 0) {
      var cellMetadata = columnMetadataMap[lastMeasuredColumnIndex];
      totalSizeOfMeasuredRows = cellMetadata.offset + cellMetadata.size;
    }

    var numUnmeasuredCells = columnCount - lastMeasuredColumnIndex - 1;
    var totalSizeOfUnmeasuredCells = numUnmeasuredCells * estimatedColumnWidth;

    return totalSizeOfMeasuredRows + totalSizeOfUnmeasuredCells;
  },

  getOffsetForColumnAndAlignment: function getOffsetForColumnAndAlignment(props, index, align, scrollOffset, instanceProps) {
    return getOffsetForIndexAndAlignment('column', props, index, align, scrollOffset, instanceProps);
  },

  getOffsetForRowAndAlignment: function getOffsetForRowAndAlignment(props, index, align, scrollOffset, instanceProps) {
    return getOffsetForIndexAndAlignment('row', props, index, align, scrollOffset, instanceProps);
  },

  getRowOffset: function getRowOffset(props, index, instanceProps) {
    return getCellMetadata('row', props, index, instanceProps).offset;
  },

  getRowHeight: function getRowHeight(props, index, instanceProps) {
    return instanceProps.rowMetadataMap[index].size;
  },

  getRowStartIndexForOffset: function getRowStartIndexForOffset(props, scrollTop, instanceProps) {
    return findNearestCell('row', props, instanceProps, scrollTop);
  },

  getRowStopIndexForStartIndex: function getRowStopIndexForStartIndex(props, startIndex, scrollTop, instanceProps) {
    var rowCount = props.rowCount,
        height = props.height;


    var cellMetadata = getCellMetadata('row', props, startIndex, instanceProps);
    var maxOffset = scrollTop + height;

    var offset = cellMetadata.offset + cellMetadata.size;
    var stopIndex = startIndex;

    while (stopIndex < rowCount - 1 && offset < maxOffset) {
      stopIndex++;
      offset += getCellMetadata('row', props, stopIndex, instanceProps).size;
    }

    return stopIndex;
  },

  initInstanceProps: function initInstanceProps(props, instance) {
    var _ref5 = props,
        estimatedColumnWidth = _ref5.estimatedColumnWidth,
        estimatedRowHeight = _ref5.estimatedRowHeight;


    var instanceProps = {
      columnMetadataMap: {},
      estimatedColumnWidth: estimatedColumnWidth || DEFAULT_ESTIMATED_CELL_SIZE,
      estimatedRowHeight: estimatedRowHeight || DEFAULT_ESTIMATED_CELL_SIZE,
      lastMeasuredColumnIndex: -1,
      lastMeasuredRowIndex: -1,
      rowMetadataMap: {}
    };

    instance.resetAfterColumnIndex = function (index) {
      instanceProps.lastMeasuredColumnIndex = index - 1;
    };

    instance.resetAfterRowIndex = function (index) {
      instanceProps.lastMeasuredRowIndex = index - 1;
    };

    return instanceProps;
  },


  validateProps: function validateProps(_ref6) {
    var columnWidth = _ref6.columnWidth,
        rowHeight = _ref6.rowHeight;

    if (process.env.NODE_ENV !== 'production') {
      if (typeof columnWidth !== 'function') {
        throw Error('An invalid "columnWidth" prop has been specified. ' + 'Value should be a function. ' + ('"' + (columnWidth === null ? 'null' : typeof columnWidth === 'undefined' ? 'undefined' : _typeof(columnWidth)) + '" was specified.'));
      } else if (typeof rowHeight !== 'function') {
        throw Error('An invalid "rowHeight" prop has been specified. ' + 'Value should be a function. ' + ('"' + (rowHeight === null ? 'null' : typeof rowHeight === 'undefined' ? 'undefined' : _typeof(rowHeight)) + '" was specified.'));
      }
    }
  }
});

var IS_SCROLLING_DEBOUNCE_INTERVAL$1 = 150;

function createListComponent(_ref) {
  var _class, _temp2;

  var getCellOffset = _ref.getCellOffset,
      getCellSize = _ref.getCellSize,
      getEstimatedTotalSize = _ref.getEstimatedTotalSize,
      getOffsetForIndexAndAlignment = _ref.getOffsetForIndexAndAlignment,
      getStartIndexForOffset = _ref.getStartIndexForOffset,
      getStopIndexForStartIndex = _ref.getStopIndexForStartIndex,
      initInstanceProps = _ref.initInstanceProps,
      validateProps = _ref.validateProps;

  return _temp2 = _class = function (_PureComponent) {
    inherits(List, _PureComponent);

    function List() {
      var _ref2;

      var _temp, _this, _ret;

      classCallCheck(this, List);

      for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      return _ret = (_temp = (_this = possibleConstructorReturn(this, (_ref2 = List.__proto__ || Object.getPrototypeOf(List)).call.apply(_ref2, [this].concat(args))), _this), _this._cellStyleCache = {}, _this._instanceProps = initInstanceProps(_this.props, _this), _this._resetIsScrollingTimeoutId = null, _this.state = {
        isScrolling: false,
        scrollDirection: 'forward',
        scrollOffset: typeof _this.props.defaultScrollOffset === 'number' ? _this.props.defaultScrollOffset : 0
      }, _this._callOnItemsRendered = memoizeOne(function (overscanStartIndex, overscanStopIndex, visibleStartIndex, visibleStopIndex) {
        return _this.props.onItemsRendered({
          overscanStartIndex: overscanStartIndex,
          overscanStopIndex: overscanStopIndex,
          visibleStartIndex: visibleStartIndex,
          visibleStopIndex: visibleStopIndex
        });
      }), _this._callOnScroll = memoizeOne(function (scrollDirection, scrollOffset) {
        return _this.props.onScroll({
          scrollDirection: scrollDirection,
          scrollOffset: scrollOffset
        });
      }), _this._getCellStyle = function (index) {
        var direction = _this.props.direction;


        var style = void 0;
        if (_this._cellStyleCache.hasOwnProperty(index)) {
          style = _this._cellStyleCache[index];
        } else {
          _this._cellStyleCache[index] = style = {
            position: 'absolute',
            left: direction === 'horizontal' ? getCellOffset(_this.props, index, _this._instanceProps) : 0,
            top: direction === 'vertical' ? getCellOffset(_this.props, index, _this._instanceProps) : 0,
            height: direction === 'vertical' ? getCellSize(_this.props, index, _this._instanceProps) : '100%',
            width: direction === 'horizontal' ? getCellSize(_this.props, index, _this._instanceProps) : '100%'
          };
        }

        return style;
      }, _this._onScrollHorizontal = function (event) {
        var scrollLeft = event.currentTarget.scrollLeft;

        _this.setState(function (prevState) {
          return {
            isScrolling: true,
            scrollDirection: prevState.scrollOffset < scrollLeft ? 'forward' : 'backward',
            scrollOffset: scrollLeft
          };
        }, _this._resetIsScrollingDebounced);
      }, _this._onScrollVertical = function (event) {
        var scrollTop = event.currentTarget.scrollTop;

        _this.setState(function (prevState) {
          return {
            isScrolling: true,
            scrollDirection: prevState.scrollOffset < scrollTop ? 'forward' : 'backward',
            scrollOffset: scrollTop
          };
        }, _this._resetIsScrollingDebounced);
      }, _this._scrollingContainerRef = function (ref) {
        _this._scrollingContainer = ref;
      }, _this._renderFunction = function (params) {
        return _this.props.children(params);
      }, _this._resetIsScrollingDebounced = function () {
        if (_this._resetIsScrollingTimeoutId !== null) {
          clearTimeout(_this._resetIsScrollingTimeoutId);
        }

        _this._resetIsScrollingTimeoutId = setTimeout(_this._resetIsScrolling, IS_SCROLLING_DEBOUNCE_INTERVAL$1);
      }, _this._resetIsScrolling = function () {
        _this._resetIsScrollingTimeoutId = null;

        _this.setState({ isScrolling: false }, function () {
          // Clear style cache after state update has been committed.
          // This way we don't break pure sCU for cells that don't use isScrolling param.
          _this._cellStyleCache = {};
        });
      }, _temp), possibleConstructorReturn(_this, _ret);
    }

    createClass(List, [{
      key: 'scrollTo',
      value: function scrollTo(scrollOffset) {
        if (this._scrollingContainer != null) {
          var _direction = this.props.direction;


          if (_direction === 'horizontal') {
            this._scrollingContainer.scrollLeft = scrollOffset;
          } else {
            this._scrollingContainer.scrollTop = scrollOffset;
          }
        }

        if (process.env.NODE_ENV === 'test') {
          // Setting scroll offset doesn't fire the onScroll callback for react-test-renderer.
          // This test-only code makes it easier to test simualted scrolling behavior.
          // It should be stripped out of any non-test code.
          if (this.props.direction === 'horizontal') {
            this._onScrollHorizontal({ currentTarget: { scrollLeft: scrollOffset } });
          } else {
            this._onScrollVertical({ currentTarget: { scrollTop: scrollOffset } });
          }
        }
      }
    }, {
      key: 'scrollToItem',
      value: function scrollToItem(index) {
        var align = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 'auto';
        var scrollOffset = this.state.scrollOffset;

        this.scrollTo(getOffsetForIndexAndAlignment(this.props, index, align, scrollOffset, this._instanceProps));
      }
    }, {
      key: 'componentDidMount',
      value: function componentDidMount() {
        var _props = this.props,
            defaultScrollOffset = _props.defaultScrollOffset,
            direction = _props.direction;


        if (typeof defaultScrollOffset === 'number') {
          if (this._scrollingContainer != null) {
            if (direction === 'horizontal') {
              this._scrollingContainer.scrollLeft = defaultScrollOffset;
            } else {
              this._scrollingContainer.scrollTop = defaultScrollOffset;
            }
          }
        }

        this._callPropsCallbacks();
      }
    }, {
      key: 'componentDidUpdate',
      value: function componentDidUpdate() {
        this._callPropsCallbacks();
      }
    }, {
      key: 'componentWillUnmount',
      value: function componentWillUnmount() {
        if (this._resetIsScrollingTimeoutId !== null) {
          clearTimeout(this._resetIsScrollingTimeoutId);
        }
      }
    }, {
      key: 'render',
      value: function render() {
        var _props2 = this.props,
            className = _props2.className,
            count = _props2.count,
            direction = _props2.direction,
            height = _props2.height,
            style = _props2.style,
            useIsScrolling = _props2.useIsScrolling,
            width = _props2.width;
        var isScrolling = this.state.isScrolling;


        var onScroll = direction === 'vertical' ? this._onScrollVertical : this._onScrollHorizontal;

        var estimatedTotalSize = getEstimatedTotalSize(this.props, this._instanceProps);

        var _getRangeToRender2 = this._getRangeToRender(),
            _getRangeToRender3 = slicedToArray(_getRangeToRender2, 2),
            startIndex = _getRangeToRender3[0],
            stopIndex = _getRangeToRender3[1];

        return React__default.createElement(
          'div',
          {
            className: className,
            ref: this._scrollingContainerRef,
            style: _extends({
              position: 'relative',
              height: height,
              width: width,
              overflow: 'auto',
              WebkitOverflowScrolling: 'touch',
              willChange: 'transform'
            }, style),
            onScroll: onScroll
          },
          React__default.createElement(
            'div',
            {
              style: {
                height: direction === 'horizontal' ? height : estimatedTotalSize,
                overflow: 'hidden',
                pointerEvents: isScrolling ? 'none' : '',
                width: direction === 'horizontal' ? estimatedTotalSize : width
              }
            },
            count > 0 && React__default.createElement(ListItems, {
              getCellStyle: this._getCellStyle,
              isScrolling: useIsScrolling ? isScrolling : undefined,
              renderFunction: this._renderFunction,
              startIndex: startIndex,
              stopIndex: stopIndex
            })
          )
        );
      }
    }, {
      key: '_callPropsCallbacks',
      value: function _callPropsCallbacks() {
        if (typeof this.props.onItemsRendered === 'function') {
          var _getRangeToRender4 = this._getRangeToRender(),
              _getRangeToRender5 = slicedToArray(_getRangeToRender4, 4),
              _overscanStartIndex = _getRangeToRender5[0],
              _overscanStopIndex = _getRangeToRender5[1],
              _visibleStartIndex = _getRangeToRender5[2],
              _visibleStopIndex = _getRangeToRender5[3];

          this._callOnItemsRendered(_overscanStartIndex, _overscanStopIndex, _visibleStartIndex, _visibleStopIndex);
        }

        if (typeof this.props.onScroll === 'function') {
          var _state = this.state,
              _scrollDirection = _state.scrollDirection,
              _scrollOffset = _state.scrollOffset;

          this._callOnScroll(_scrollDirection, _scrollOffset);
        }
      }

      // Lazily create and cache cell styles while scrolling,
      // So that pure component sCU will prevent re-renders.

    }, {
      key: '_getRangeToRender',
      value: function _getRangeToRender() {
        var _props3 = this.props,
            count = _props3.count,
            overscanCount = _props3.overscanCount;
        var _state2 = this.state,
            scrollDirection = _state2.scrollDirection,
            scrollOffset = _state2.scrollOffset;


        var startIndex = getStartIndexForOffset(this.props, scrollOffset, this._instanceProps);
        var stopIndex = getStopIndexForStartIndex(this.props, startIndex, scrollOffset, this._instanceProps);

        // Overscan by one cell in each direction so that tab/focus works.
        // If there isn't at least one extra cell, tab loops back around.
        var overscanBackward = scrollDirection === 'backward' ? Math.max(1, overscanCount) : 1;
        var overscanForward = scrollDirection === 'forward' ? Math.max(1, overscanCount) : 1;

        return [Math.max(0, startIndex - overscanBackward), Math.max(0, Math.min(count - 1, stopIndex + overscanForward)), startIndex, stopIndex];
      }

      // Facade for the user-provided children function.
      // This adds the overhead of an additional method call,
      // But has hte benefit of not breaking pure sCU checks,
      // Allowing List to avoid re-rendering cells until indices change.

    }], [{
      key: 'getDerivedStateFromProps',
      value: function getDerivedStateFromProps(nextProps, prevState) {
        validateSharedProps$1(nextProps);
        validateProps(nextProps);
        return null;
      }
    }]);
    return List;
  }(React.PureComponent), _class.defaultProps = {
    direction: 'vertical',
    overscanCount: 2,
    useIsScrolling: false
  }, _temp2;
}

var ListItems = function (_PureComponent2) {
  inherits(ListItems, _PureComponent2);

  function ListItems() {
    classCallCheck(this, ListItems);
    return possibleConstructorReturn(this, (ListItems.__proto__ || Object.getPrototypeOf(ListItems)).apply(this, arguments));
  }

  createClass(ListItems, [{
    key: 'render',
    value: function render() {
      var _props4 = this.props,
          getCellStyle = _props4.getCellStyle,
          isScrolling = _props4.isScrolling,
          renderFunction = _props4.renderFunction,
          startIndex = _props4.startIndex,
          stopIndex = _props4.stopIndex;


      var cells = [];

      for (var _index = startIndex; _index <= stopIndex; _index++) {
        var _key2 = '' + _index;
        var _style = getCellStyle(_index);

        cells.push(renderFunction({
          key: _key2,
          index: _index,
          isScrolling: isScrolling,
          style: _style
        }));
      }

      return cells;
    }
  }]);
  return ListItems;
}(React.PureComponent);

// NOTE: I considered further wrapping individual items with a pure ListItem component.
// This would avoid ever calling the render function for the same index more than once,
// But it would also add the overhead of a lot of components/fibers.
// I assume people already do this (render function returning a class component),
// So my doing it would just unnecessarily double the wrappers.

var validateSharedProps$1 = function validateSharedProps(_ref3) {
  var children = _ref3.children,
      direction = _ref3.direction,
      height = _ref3.height,
      width = _ref3.width;

  if (process.env.NODE_ENV !== 'production') {
    if (direction !== 'horizontal' && direction !== 'vertical') {
      throw Error('An invalid "direction" prop has been specified. ' + 'Value should be either "horizontal" or "vertical". ' + ('"' + direction + '" was specified.'));
    }

    if (typeof children !== 'function') {
      throw Error('An invalid "children" prop has been specified. ' + 'Value should be a function that creates a React element. ' + ('"' + (children === null ? 'null' : typeof children === 'undefined' ? 'undefined' : _typeof(children)) + '" was specified.'));
    }

    if (direction === 'horizontal' && typeof width !== 'number') {
      throw Error('An invalid "width" prop has been specified. ' + 'Horizontal lists must specify a number for width. ' + ('"' + (width === null ? 'null' : typeof width === 'undefined' ? 'undefined' : _typeof(width)) + '" was specified.'));
    } else if (direction === 'vertical' && typeof height !== 'number') {
      throw Error('An invalid "height" prop has been specified. ' + 'Vertical lists must specify a number for height. ' + ('"' + (height === null ? 'null' : typeof height === 'undefined' ? 'undefined' : _typeof(height)) + '" was specified.'));
    }
  }
};

var DEFAULT_ESTIMATED_CELL_SIZE$1 = 50;

var getCellMetadata$1 = function getCellMetadata(props, index, instanceProps) {
  var _ref = props,
      cellSize = _ref.cellSize;
  var cellMetadataMap = instanceProps.cellMetadataMap,
      lastMeasuredIndex = instanceProps.lastMeasuredIndex;


  if (index > lastMeasuredIndex) {
    var _offset = 0;
    if (lastMeasuredIndex >= 0) {
      var cellMetadata = cellMetadataMap[lastMeasuredIndex];
      _offset = cellMetadata.offset + cellMetadata.size;
    }

    for (var i = lastMeasuredIndex + 1; i <= index; i++) {
      var _size = cellSize(i);

      cellMetadataMap[i] = {
        offset: _offset,
        size: _size
      };

      _offset += _size;
    }

    instanceProps.lastMeasuredIndex = index;
  }

  return cellMetadataMap[index];
};

var findNearestCell$1 = function findNearestCell(props, instanceProps, offset) {
  var cellMetadataMap = instanceProps.cellMetadataMap,
      lastMeasuredIndex = instanceProps.lastMeasuredIndex;


  var lastMeasuredCellOffset = lastMeasuredIndex > 0 ? cellMetadataMap[lastMeasuredIndex].offset : 0;

  if (lastMeasuredCellOffset >= offset) {
    // If we've already measured cells within this range just use a binary search as it's faster.
    return findNearestCellBinarySearch$1(props, instanceProps, lastMeasuredIndex, 0, offset);
  } else {
    // If we haven't yet measured this high, fallback to an exponential search with an inner binary search.
    // The exponential search avoids pre-computing sizes for the full set of cells as a binary search would.
    // The overall complexity for this approach is O(log n).
    return findNearestCellExponentialSearch$1(props, instanceProps, lastMeasuredIndex, offset);
  }
};

var findNearestCellBinarySearch$1 = function findNearestCellBinarySearch(props, instanceProps, high, low, offset) {
  while (low <= high) {
    var middle = low + Math.floor((high - low) / 2);
    var currentOffset = getCellMetadata$1(props, middle, instanceProps).offset;

    if (currentOffset === offset) {
      return middle;
    } else if (currentOffset < offset) {
      low = middle + 1;
    } else if (currentOffset > offset) {
      high = middle - 1;
    }
  }

  if (low > 0) {
    return low - 1;
  } else {
    return 0;
  }
};

var findNearestCellExponentialSearch$1 = function findNearestCellExponentialSearch(props, instanceProps, index, offset) {
  var count = props.count;

  var interval = 1;

  while (index < count && getCellMetadata$1(props, index, instanceProps).offset < offset) {
    index += interval;
    interval *= 2;
  }

  return findNearestCellBinarySearch$1(props, instanceProps, Math.min(index, count - 1), Math.floor(index / 2), offset);
};

var getEstimatedTotalSize = function getEstimatedTotalSize(_ref2, _ref3) {
  var count = _ref2.count;
  var cellMetadataMap = _ref3.cellMetadataMap,
      estimatedCellSize = _ref3.estimatedCellSize,
      lastMeasuredIndex = _ref3.lastMeasuredIndex;

  var totalSizeOfMeasuredCells = 0;

  if (lastMeasuredIndex >= 0) {
    var cellMetadata = cellMetadataMap[lastMeasuredIndex];
    totalSizeOfMeasuredCells = cellMetadata.offset + cellMetadata.size;
  }

  var numUnmeasuredCells = count - lastMeasuredIndex - 1;
  var totalSizeOfUnmeasuredCells = numUnmeasuredCells * estimatedCellSize;

  return totalSizeOfMeasuredCells + totalSizeOfUnmeasuredCells;
};

var VariableSizeList = createListComponent({
  getCellOffset: function getCellOffset(props, index, instanceProps) {
    return getCellMetadata$1(props, index, instanceProps).offset;
  },

  getCellSize: function getCellSize(props, index, instanceProps) {
    return instanceProps.cellMetadataMap[index].size;
  },

  getEstimatedTotalSize: getEstimatedTotalSize,

  getOffsetForIndexAndAlignment: function getOffsetForIndexAndAlignment(props, index, align, scrollOffset, instanceProps) {
    var direction = props.direction,
        height = props.height,
        width = props.width;


    var size = direction === 'horizontal' ? width : height;
    var cellMetadata = getCellMetadata$1(props, index, instanceProps);

    // Get estimated total size after CellMetadata is computed,
    // To ensure it reflects actual measurements instead of just estimates.
    var estimatedTotalSize = getEstimatedTotalSize(props, instanceProps);

    var maxOffset = Math.min(estimatedTotalSize - size, cellMetadata.offset);
    var minOffset = Math.max(0, cellMetadata.offset - size + cellMetadata.size);

    switch (align) {
      case 'start':
        return maxOffset;
      case 'end':
        return minOffset;
      case 'center':
        return Math.round(minOffset + (maxOffset - minOffset) / 2);
      case 'auto':
      default:
        if (scrollOffset >= minOffset && scrollOffset <= maxOffset) {
          return scrollOffset;
        } else if (scrollOffset - minOffset < maxOffset - scrollOffset) {
          return minOffset;
        } else {
          return maxOffset;
        }
    }
  },

  getStartIndexForOffset: function getStartIndexForOffset(props, offset, instanceProps) {
    return findNearestCell$1(props, instanceProps, offset);
  },

  getStopIndexForStartIndex: function getStopIndexForStartIndex(props, startIndex, scrollOffset, instanceProps) {
    var count = props.count,
        direction = props.direction,
        height = props.height,
        width = props.width;


    var size = direction === 'horizontal' ? width : height;
    var cellMetadata = getCellMetadata$1(props, startIndex, instanceProps);
    var maxOffset = scrollOffset + size;

    var offset = cellMetadata.offset + cellMetadata.size;
    var stopIndex = startIndex;

    while (stopIndex < count - 1 && offset < maxOffset) {
      stopIndex++;
      offset += getCellMetadata$1(props, stopIndex, instanceProps).size;
    }

    return stopIndex;
  },

  initInstanceProps: function initInstanceProps(props, instance) {
    var _ref4 = props,
        estimatedCellSize = _ref4.estimatedCellSize;


    var instanceProps = {
      cellMetadataMap: {},
      estimatedCellSize: estimatedCellSize || DEFAULT_ESTIMATED_CELL_SIZE$1,
      lastMeasuredIndex: -1
    };

    instance.resetAfterIndex = function (index) {
      instanceProps.lastMeasuredIndex = index - 1;
    };

    return instanceProps;
  },


  validateProps: function validateProps(_ref5) {
    var cellSize = _ref5.cellSize;

    if (process.env.NODE_ENV !== 'production') {
      if (typeof cellSize !== 'function') {
        throw Error('An invalid "cellSize" prop has been specified. ' + 'Value should be a function. ' + ('"' + (cellSize === null ? 'null' : typeof cellSize === 'undefined' ? 'undefined' : _typeof(cellSize)) + '" was specified.'));
      }
    }
  }
});

var FixedSizeGrid = createGridComponent({
  getColumnOffset: function getColumnOffset(_ref, index) {
    var columnWidth = _ref.columnWidth;
    return index * columnWidth;
  },

  getColumnWidth: function getColumnWidth(_ref2, index) {
    var columnWidth = _ref2.columnWidth;
    return columnWidth;
  },

  getRowOffset: function getRowOffset(_ref3, index) {
    var rowHeight = _ref3.rowHeight;
    return index * rowHeight;
  },

  getRowHeight: function getRowHeight(_ref4, index) {
    var rowHeight = _ref4.rowHeight;
    return rowHeight;
  },

  getEstimatedTotalHeight: function getEstimatedTotalHeight(_ref5) {
    var rowCount = _ref5.rowCount,
        rowHeight = _ref5.rowHeight;
    return rowHeight * rowCount;
  },

  getEstimatedTotalWidth: function getEstimatedTotalWidth(_ref6) {
    var columnCount = _ref6.columnCount,
        columnWidth = _ref6.columnWidth;
    return columnWidth * columnCount;
  },

  getOffsetForColumnAndAlignment: function getOffsetForColumnAndAlignment(_ref7, columnIndex, align, scrollLeft) {
    var columnCount = _ref7.columnCount,
        columnWidth = _ref7.columnWidth,
        width = _ref7.width;

    var maxOffset = Math.min(columnCount * columnWidth - width, columnIndex * columnWidth);
    var minOffset = Math.max(0, columnIndex * columnWidth - width + columnWidth);

    switch (align) {
      case 'start':
        return maxOffset;
      case 'end':
        return minOffset;
      case 'center':
        return Math.round(minOffset + (maxOffset - minOffset) / 2);
      case 'auto':
      default:
        if (scrollLeft >= minOffset && scrollLeft <= maxOffset) {
          return scrollLeft;
        } else if (scrollLeft - minOffset < maxOffset - scrollLeft) {
          return minOffset;
        } else {
          return maxOffset;
        }
    }
  },

  getOffsetForRowAndAlignment: function getOffsetForRowAndAlignment(_ref8, rowIndex, align, scrollTop) {
    var rowHeight = _ref8.rowHeight,
        height = _ref8.height,
        rowCount = _ref8.rowCount;

    var maxOffset = Math.min(rowCount * rowHeight - height, rowIndex * rowHeight);
    var minOffset = Math.max(0, rowIndex * rowHeight - height + rowHeight);

    switch (align) {
      case 'start':
        return maxOffset;
      case 'end':
        return minOffset;
      case 'center':
        return Math.round(minOffset + (maxOffset - minOffset) / 2);
      case 'auto':
      default:
        if (scrollTop >= minOffset && scrollTop <= maxOffset) {
          return scrollTop;
        } else if (scrollTop - minOffset < maxOffset - scrollTop) {
          return minOffset;
        } else {
          return maxOffset;
        }
    }
  },

  getColumnStartIndexForOffset: function getColumnStartIndexForOffset(_ref9, scrollLeft) {
    var columnWidth = _ref9.columnWidth,
        columnCount = _ref9.columnCount;
    return Math.max(0, Math.min(columnCount - 1, Math.floor(scrollLeft / columnWidth)));
  },

  getColumnStopIndexForStartIndex: function getColumnStopIndexForStartIndex(_ref10, startIndex, scrollLeft) {
    var columnWidth = _ref10.columnWidth,
        columnCount = _ref10.columnCount,
        width = _ref10.width;

    var left = startIndex * columnWidth;
    return Math.max(0, Math.min(columnCount - 1, startIndex + Math.floor((width + (scrollLeft - left)) / columnWidth)));
  },

  getRowStartIndexForOffset: function getRowStartIndexForOffset(_ref11, scrollTop) {
    var rowHeight = _ref11.rowHeight,
        rowCount = _ref11.rowCount;
    return Math.max(0, Math.min(rowCount - 1, Math.floor(scrollTop / rowHeight)));
  },

  getRowStopIndexForStartIndex: function getRowStopIndexForStartIndex(_ref12, startIndex, scrollTop) {
    var rowHeight = _ref12.rowHeight,
        rowCount = _ref12.rowCount,
        height = _ref12.height;

    var left = startIndex * rowHeight;
    return Math.max(0, Math.min(rowCount - 1, startIndex + Math.floor((height + (scrollTop - left)) / rowHeight)));
  },

  initInstanceProps: function initInstanceProps(props) {
    // Noop
  },


  validateProps: function validateProps(_ref13) {
    var columnWidth = _ref13.columnWidth,
        rowHeight = _ref13.rowHeight;

    if (process.env.NODE_ENV !== 'production') {
      if (typeof columnWidth !== 'number') {
        throw Error('An invalid "columnWidth" prop has been specified. ' + 'Value should be a number. ' + ('"' + (columnWidth === null ? 'null' : typeof columnWidth === 'undefined' ? 'undefined' : _typeof(columnWidth)) + '" was specified.'));
      }

      if (typeof rowHeight !== 'number') {
        throw Error('An invalid "rowHeight" prop has been specified. ' + 'Value should be a number. ' + ('"' + (rowHeight === null ? 'null' : typeof rowHeight === 'undefined' ? 'undefined' : _typeof(rowHeight)) + '" was specified.'));
      }
    }
  }
});

var FixedSizeList = createListComponent({
  getCellOffset: function getCellOffset(_ref, index) {
    var cellSize = _ref.cellSize,
        size = _ref.size;
    return index * cellSize;
  },

  getCellSize: function getCellSize(_ref2, index) {
    var cellSize = _ref2.cellSize,
        size = _ref2.size;
    return cellSize;
  },

  getEstimatedTotalSize: function getEstimatedTotalSize(_ref3) {
    var cellSize = _ref3.cellSize,
        count = _ref3.count;
    return cellSize * count;
  },

  getOffsetForIndexAndAlignment: function getOffsetForIndexAndAlignment(_ref4, index, align, scrollOffset) {
    var cellSize = _ref4.cellSize,
        count = _ref4.count,
        direction = _ref4.direction,
        height = _ref4.height,
        width = _ref4.width;

    var size = direction === 'horizontal' ? width : height;
    var maxOffset = Math.min(count * cellSize - size, index * cellSize);
    var minOffset = Math.max(0, index * cellSize - size + cellSize);

    switch (align) {
      case 'start':
        return maxOffset;
      case 'end':
        return minOffset;
      case 'center':
        return Math.round(minOffset + (maxOffset - minOffset) / 2);
      case 'auto':
      default:
        if (scrollOffset >= minOffset && scrollOffset <= maxOffset) {
          return scrollOffset;
        } else if (scrollOffset - minOffset < maxOffset - scrollOffset) {
          return minOffset;
        } else {
          return maxOffset;
        }
    }
  },

  getStartIndexForOffset: function getStartIndexForOffset(_ref5, offset) {
    var cellSize = _ref5.cellSize,
        count = _ref5.count;
    return Math.max(0, Math.min(count - 1, Math.floor(offset / cellSize)));
  },

  getStopIndexForStartIndex: function getStopIndexForStartIndex(_ref6, startIndex, scrollOffset) {
    var cellSize = _ref6.cellSize,
        count = _ref6.count,
        direction = _ref6.direction,
        height = _ref6.height,
        width = _ref6.width;

    var offset = startIndex * cellSize;
    var size = direction === 'horizontal' ? width : height;
    return Math.max(0, Math.min(count - 1, startIndex + Math.floor((size + (scrollOffset - offset)) / cellSize)));
  },

  initInstanceProps: function initInstanceProps(props) {
    // Noop
  },


  validateProps: function validateProps(_ref7) {
    var cellSize = _ref7.cellSize;

    if (process.env.NODE_ENV !== 'production') {
      if (typeof cellSize !== 'number') {
        throw Error('An invalid "cellSize" prop has been specified. ' + 'Value should be a number. ' + ('"' + (cellSize === null ? 'null' : typeof cellSize === 'undefined' ? 'undefined' : _typeof(cellSize)) + '" was specified.'));
      }
    }
  }
});

exports.VariableSizeGrid = VariableSizeGrid;
exports.VariableSizeList = VariableSizeList;
exports.FixedSizeGrid = FixedSizeGrid;
exports.FixedSizeList = FixedSizeList;
